// BLOGGIE UNWARPER (For Still Panoramic Images)
// By Golan Levin, June 2010
// http://www.flong.com
// For Processing 1.1 or 0185 

//=============================================
// Important parameters for the unwarping
float imgCx; // optical x-center of the warped (annular) image
float imgCy; // optical y-center of the warped (annular) image 
float maxR;  // outer radius of the warped annulus, in pixels
float minR;  // inner radius of the warped annulus, in pixels

// The input (warped, annular) image
PImage inputImage;
int inputImageW, inputImageH;
float inputImageScalingFactor;
float inputImageDrawingWidth;
float inputImageDrawingHeight;

// The output (unwarped, panoramic) image
PImage outputImage;
float  unwarpedAspectRatio;
int    unwarpedW, unwarpedH;
float  outputImageDrawingHeight;

// Intermediate data structures for the unwarping
float srcxArray[];
float srcyArray[];
int patchR[][];
int patchG[][];
int patchB[][];

// Interaction state variables
boolean bMousePressedInInputImage;
boolean bMousePressedInCylinderRegion;
boolean bMousePressedInOutputImage;
boolean bUseCubicInterpolation; 
float   cylinderRotateX = 0;
float   cylinderRotateY = 0;

// Don't touch these parameters, 
// they are specific to the nonlinearities of the Bloggie lens.
float yWarpA =   0.1850;
float yWarpB =   0.8184;
float yWarpC =  -0.0028;
float angularShift = 0;


//=============================================
void setup() {
  size(1200,620, P3D);

  inputImage = loadImage("input.jpg"); 
  inputImageW = inputImage.width;
  inputImageH = inputImage.height;

  maxR  = inputImageH/2.0 * 0.72;
  minR  = inputImageH/2.0 * 0.16;
  imgCx = 1314.5; //inputImageW/2;
  imgCy = 997.25; //inputImageH/2;

  // The Bloggie has a 55-degree vertical FOV, and a 360-degree horizontal FOV.
  unwarpedAspectRatio = 360.0/55.0;
  unwarpedW = width;
  unwarpedH = (int)(unwarpedW / unwarpedAspectRatio); 
  outputImageDrawingHeight = height-unwarpedH;
  outputImage = createImage(unwarpedW, unwarpedH, RGB);

  srcxArray = new float[unwarpedW*unwarpedH];
  srcyArray = new float[unwarpedW*unwarpedH];
  patchR = new int[4][4];
  patchG = new int[4][4];
  patchB = new int[4][4];

  inputImageScalingFactor  = outputImageDrawingHeight / (float)inputImageH;
  inputImageDrawingWidth   = inputImageW * inputImageScalingFactor;
  inputImageDrawingHeight  = inputImageH * inputImageScalingFactor;

  bUseCubicInterpolation = false;
  computeInversePolarTransform();
}


//=============================================
void draw() {
  background(64,64,64);

  computeInversePolarTransform();
  
  // slow but smooth OR fast but inaccurate and pixelated
  if (bUseCubicInterpolation) {
    unwarpBicubic();
  } else {
    unwarpNearestNeighbor();
  }

  drawTexturedCylinder();
  drawOutputImage();
  drawInputImage();
}


//=============================================
void keyPressed() {
  switch (key) {
  case 'c': 
    // switch between cubic and nearest-neighbor interpolation.
    bUseCubicInterpolation = !bUseCubicInterpolation; 
    break;
  case 'e':
    // export the unwarped output image.
    outputImage.save("output.jpg");
    break;
  }

  // OR, nudge the center point for unwarping the input image.
  float incr = 0.25;
  switch (keyCode) {
  case 37: // LEFT
    imgCx -= incr;
    println ("Input image center: (" + imgCx + ", " + imgCy + ")");
    break;
  case 39: // RIGHT
    imgCx += incr;
    println ("Input image center: (" + imgCx + ", " + imgCy + ")");
    break;
  case 40: // DOWN
    imgCy += incr;
    println ("Input image center: (" + imgCx + ", " + imgCy + ")");
    break;
  case 38: // UP
    imgCy -= incr;
    println ("Input image center: (" + imgCx + ", " + imgCy + ")");
    break;
  }
}


//=============================================
void mouseReleased() {
  bMousePressedInInputImage     = false;
  bMousePressedInCylinderRegion = false;
  bMousePressedInOutputImage    = false;
}

void mousePressed() {
  bMousePressedInInputImage     = false;
  bMousePressedInCylinderRegion = false;
  bMousePressedInOutputImage    = false;

  if ((mouseX < inputImageDrawingWidth) && (mouseY < inputImageDrawingHeight)) {
    bMousePressedInInputImage = true;
  } 
  else if ((mouseX > inputImageDrawingWidth) && (mouseY < inputImageDrawingHeight)) {
    bMousePressedInCylinderRegion = true;
  } 
  else if (mouseY > inputImageDrawingHeight) {
    bMousePressedInOutputImage = true;
  }
}

void mouseDragged() {
  if (bMousePressedInInputImage && (mouseX < inputImageDrawingWidth) && (mouseY < inputImageDrawingHeight)) {
    imgCx = (float)mouseX/(float)inputImageDrawingWidth  * inputImageW;
    imgCy = (float)mouseY/(float)inputImageDrawingHeight * inputImageH;
    println ("Input image center: (" + imgCx + ", " + imgCy + ")");
  } 
  else if (bMousePressedInCylinderRegion) {
    cylinderRotateX = map(mouseY, 0, height, -PI, PI);
    cylinderRotateY = map(mouseX, 0, width,  -PI, PI);
  } 
  else if (bMousePressedInOutputImage && (mouseY > inputImageDrawingHeight)) {
    angularShift = map (mouseX,0,width, 0, TWO_PI);
  }
}

//=============================================
void drawOutputImage() {
  // draw the unwarped panoramic strip
  image(outputImage, 0, outputImageDrawingHeight);
}

//=============================================
void drawInputImage() {
  // draw the warped input image
  image (inputImage, 0,0, inputImageDrawingWidth, inputImageDrawingHeight);

  // draw crosshairs at unwarping center point
  smooth();
  stroke(255,0,0);
  float crossSize = 8;
  float crossX = map(imgCx, 0, inputImageW, 0, inputImageDrawingWidth); 
  float crossY = map(imgCy, 0, inputImageH, 0, inputImageDrawingHeight); 
  line (crossX, crossY-crossSize, crossX, crossY+crossSize);
  line (crossX-crossSize, crossY, crossX+crossSize, crossY);
  noSmooth();
}

//=============================================
void drawTexturedCylinder() {

  float cylx = inputImageDrawingWidth + (width-inputImageDrawingWidth)/2.0;
  float cyly = inputImageDrawingHeight/2.0;
  float cylr = 100.0 * 1.5;
  float cylh =  57.0 * 1.5;

  pushMatrix();
  translate(cylx, cyly, 0);
  rotateX(cylinderRotateX);
  rotateY(cylinderRotateY);

  noStroke();
  noSmooth();
  fill(255,255,255);
  beginShape(QUAD_STRIP);
  texture(outputImage);

  int tubeRes = 90;
  float angle = 360.0 / (tubeRes);
  for (int i = 0; i <= tubeRes; i++) {

    float tubeX = cos(radians(i * angle));
    float tubeY = sin(radians(i * angle));

    float x = tubeX * cylr;
    float z = tubeY * cylr;
    float u = (outputImage.width / (float)tubeRes) * (float) i;
    vertex(x, -cylh, z, u, 0);
    vertex(x,  cylh, z, u, outputImage.height);
  }
  endShape();
  popMatrix();
}



//=============================================
void computeInversePolarTransform() {
  int   srcx, srcy, srcIndex;
  int   dstRow, dstIndex;
  float radius, angle;

  for (int dsty=0; dsty<unwarpedH; dsty++) {
    float y = ((float)dsty/(float)unwarpedH); //0..1
    float yfrac = yWarpA*y*y + yWarpB*y + yWarpC;
    radius  = (yfrac * (maxR-minR)) + minR; 
    dstRow = dsty*unwarpedW;

    for (int dstx=0; dstx<unwarpedW; dstx++) {
      dstIndex = dstRow + dstx;
      angle    = (0 - ((float)dstx/(float)unwarpedW) * TWO_PI) + angularShift;

      srcxArray[dstIndex] = imgCx + radius*cos(angle);
      srcyArray[dstIndex] = imgCy + radius*sin(angle);
    }
  }
}



//=============================================
void unwarpNearestNeighbor() {
  
  int inputImageN = inputImageW * inputImageH; 
  int dstRow, dstIndex;
  int srcx, srcy, srcIndex;
  color black = color(0,0,0); 
  color col;
  
  outputImage.loadPixels();
  for (int dsty=0; dsty<unwarpedH; dsty++) {
    dstRow = dsty*unwarpedW;

    for (int dstx=0; dstx<unwarpedW; dstx++) {
      dstIndex = dstRow + dstx;
      srcx  = (int) srcxArray[dstIndex];
      srcy  = (int) srcyArray[dstIndex];
      srcIndex = srcy*inputImageW + srcx;

      col = black;
      if ((srcIndex >= 0) && (srcIndex < inputImageN)) {
        col = inputImage.pixels[srcIndex];
      }
      outputImage.pixels[dstIndex] = col;
    }
  }
  outputImage.updatePixels();
}



//=============================================
void unwarpBicubic() {
  int dstRow, dstIndex;
  float srcxf, srcyf;
  float px, py;
  float px2, py2;
  float px3, py3; 

  float interpR, interpG, interpB;
  int srcx, srcy, srcIndex;
  color black = color(0,0,0); 
  color srcColor;
  color col;
  int patchIndex;
  int loIndex = inputImageW+1;
  int hiIndex = (inputImageW*inputImageH)-(inputImageW*3)-1;
  int patchRow;
  color inputPixels[] = inputImage.pixels;

  outputImage.loadPixels();
  for (int dsty=0; dsty<unwarpedH; dsty++) {
    dstRow = dsty*width;

    for (int dstx=0; dstx<unwarpedW; dstx++) {
      dstIndex = dstRow + dstx;
      srcxf = srcxArray[dstIndex];
      srcyf = srcyArray[dstIndex];
      srcx  = (int) srcxf;
      srcy  = (int) srcyf;
      srcIndex = srcy*inputImageW + srcx;

      col = black;
      srcColor = black;

      for (int dy=0; dy<4; dy++) {
        patchRow = srcIndex + ((dy-1)*inputImageW);
        for (int dx=0; dx<4; dx++) {
          patchIndex = patchRow + (dx-1);
          if ((patchIndex >= loIndex) && (patchIndex < hiIndex)) {
            srcColor = inputPixels[patchIndex];
          }
          patchR[dx][dy] = (srcColor & 0x00FF0000) >> 16;
          patchG[dx][dy] = (srcColor & 0x0000FF00) >>  8;
          patchB[dx][dy] = (srcColor & 0x000000FF)      ;
        }
      }

      px = srcxf - srcx;
      py = srcyf - srcy;
      px2 =  px * px;
      px3 = px2 * px;
      py2 =  py * py;
      py3 = py2 * py;

      interpR = bicubicInterpolate (patchR, px,py, px2,py2, px3,py3);
      interpG = bicubicInterpolate (patchG, px,py, px2,py2, px3,py3);
      interpB = bicubicInterpolate (patchB, px,py, px2,py2, px3,py3);

      col = color (interpR, interpG, interpB);
      outputImage.pixels[dstIndex] = col;
    }
  }
  outputImage.updatePixels();
}


//=============================================
float bicubicInterpolate (int[][] p, float x,float y, float x2,float y2, float x3,float y3) {
  // adapted from http://www.paulinternet.nl/?page=bicubic. 
  // Note that this code can produce values outside of 0...255, due to cubic overshoot. 
  // Processing prvents that from happening, but C++ doesnt. Clamp the output if this happens.

  int p00 = p[0][0];  
  int p10 = p[1][0]; 
  int p20 = p[2][0]; 
  int p30 = p[3][0]; 

  int p01 = p[0][1];  
  int p11 = p[1][1]; 
  int p21 = p[2][1]; 
  int p31 = p[3][1]; 

  int p02 = p[0][2];  
  int p12 = p[1][2]; 
  int p22 = p[2][2]; 
  int p32 = p[3][2]; 

  int p03 = p[0][3];  
  int p13 = p[1][3]; 
  int p23 = p[2][3]; 
  int p33 = p[3][3]; 

  int a00 =    p11;
  int a01 =   -p10 +   p12;
  int a02 =  2*p10 - 2*p11 +   p12 -   p13;
  int a03 =   -p10 +   p11 -   p12 +   p13;
  int a10 =   -p01 +   p21;
  int a11 =    p00 -   p02 -   p20 +   p22;
  int a12 = -2*p00 + 2*p01 -   p02 +   p03 + 2*p20 - 2*p21 +   p22 -   p23;
  int a13 =    p00 -   p01 +   p02 -   p03 -   p20 +   p21 -   p22 +   p23;
  int a20 =  2*p01 - 2*p11 +   p21 -   p31;
  int a21 = -2*p00 + 2*p02 + 2*p10 - 2*p12 -   p20 +   p22 +   p30 -   p32;
  int a22 =  4*p00 - 4*p01 + 2*p02 - 2*p03 - 4*p10 + 4*p11 - 2*p12 + 2*p13 + 2*p20 - 2*p21 + p22 - p23 - 2*p30 + 2*p31 - p32 + p33;
  int a23 = -2*p00 + 2*p01 - 2*p02 + 2*p03 + 2*p10 - 2*p11 + 2*p12 - 2*p13 -   p20 +   p21 - p22 + p23 +   p30 -   p31 + p32 - p33;
  int a30 =   -p01 +   p11 -   p21 +   p31; 
  int a31 =    p00 -   p02 -   p10 +   p12 +   p20 -   p22 -   p30 +   p32;
  int a32 = -2*p00 + 2*p01 -   p02 +   p03 + 2*p10 - 2*p11 +   p12 -   p13 - 2*p20 + 2*p21 - p22 + p23 + 2*p30 - 2*p31 + p32 - p33;
  int a33 =    p00 -   p01 +   p02 -   p03 -   p10 +   p11 -   p12 +   p13 +   p20 -   p21 + p22 - p23 -   p30 +   p31 - p32 + p33;

  return 
    a00      + a01 * y      + a02 * y2      + a03 * y3 +
    a10 * x  + a11 * x  * y + a12 * x  * y2 + a13 * x  * y3 +
    a20 * x2 + a21 * x2 * y + a22 * x2 * y2 + a23 * x2 * y3 +
    a30 * x3 + a31 * x3 * y + a32 * x3 * y2 + a33 * x3 * y3;
}








